from aioipfs.api import SubAPI
from aioipfs.api import ARG_PARAM
from aioipfs.api import boolarg
from aioipfs import multi


class SwarmPeeringAPI(SubAPI):
    async def add(self, peer: str):
        """
        Add peers into the peering subsystem.

        :param str peer: address of peer to add into the peering subsystem
        """

        return await self.fetch_json(
            self.url('swarm/peering/add'),
            params={ARG_PARAM: peer}
        )

    async def ls(self):
        """
        List peers registered in the peering subsystem.
        """

        return await self.fetch_json(self.url('swarm/peering/ls'))

    async def rm(self, peer: str):
        """
        Remove a peer from the peering subsystem.

        :param str peer: ID of peer to remove from the peering subsystem
        """

        return await self.fetch_json(
            self.url('swarm/peering/rm'),
            params={ARG_PARAM: peer}
        )


class SwarmAPI(SubAPI):
    def __init__(self, driver):
        super().__init__(driver)

        self.peering = SwarmPeeringAPI(driver)

    async def peers(self, verbose=True, streams=False, latency=False,
                    direction=False):
        """
        List peers with open connections.

        :param bool verbose: display all extra information
        :param bool streams: Also list information about
            open streams for each peer
        :param bool latency: Also list information about
            latency to each peer
        :param bool direction: Also list information about
            the direction of connection
        """
        params = {
            'verbose': boolarg(verbose),
            'streams': boolarg(streams),
            'latency': boolarg(latency),
            'direction': boolarg(direction)
        }

        return await self.fetch_json(self.url('swarm/peers'), params=params)

    async def addrs(self):
        """
        List known addresses. Useful for debugging.
        """
        return await self.fetch_json(self.url('swarm/addrs'))

    async def addrs_local(self, id=False):
        """
        List local addresses.
        """
        params = {'id': boolarg(id)}
        return await self.fetch_json(self.url('swarm/addrs/local'),
                                     params=params)

    async def addrs_listen(self):
        """
        List interface listening addresses.
        """
        return await self.fetch_json(self.url('swarm/addrs/listen'))

    async def connect(self, peer):
        """
        Open connection to a given address.

        :param str peer: address of peer to connect to
        """
        params = {ARG_PARAM: peer}
        return await self.fetch_json(self.url('swarm/connect'),
                                     params=params)

    async def disconnect(self, peer):
        """
        Close connection to a given address.

        :param str peer: address of peer to disconnect from
        """
        params = {ARG_PARAM: peer}
        return await self.fetch_json(self.url('swarm/disconnect'),
                                     params=params)

    async def filters_add(self, filter):
        """
        Add an address filter.

        :param str filter: Multiaddr to filter
        """
        params = {ARG_PARAM: filter}
        return await self.fetch_json(self.url('swarm/filters/add'),
                                     params=params)

    async def filters_rm(self, filter):
        """
        Remove an address filter.

        :param str filter: Multiaddr filter to remove
        """
        params = {ARG_PARAM: filter}
        return await self.fetch_json(self.url('swarm/filters/rm'),
                                     params=params)

    async def limit(self, filepath, scope: str):
        """
        Get or set resource limits for a scope

        :param str filepath: path to the (JSON) file containing the limits
        :param str scope: scope of the limit
        """
        params = {ARG_PARAM: scope}

        with multi.FormDataWriter() as mpwriter:
            mpwriter.append_payload(multi.bytes_payload_from_file(filepath))

            return await self.post(self.url('swarm/limit'), mpwriter,
                                   params=params, outformat='json')

    async def resources(self):
        """
        Get a summary of all resources accounted for by the
        libp2p Resource Manager.
        """

        return await self.fetch_json(self.url('swarm/resources'))

    async def stats(self, scope: str):
        """
        Report resource usage for a scope

        :param str scope: scope of the stat report
        """

        return await self.fetch_json(self.url('swarm/stats'),
                                     params={ARG_PARAM: scope})



import json
from functools import reduce
from pathlib import Path

try:
    import ipfs_car_decoder as car_decoder
    have_car_decoder = True
except Exception:
    have_car_decoder = False
    car_decoder = None


class CARDecoderMissing(Exception):
    pass


def car_open(car_path: Path):
    """
    Open a Content-Addressable aRchive file and return the CAR stream.

    :param Path car_path: CAR file path
    """

    if not have_car_decoder:
        raise CARDecoderMissing()

    return car_decoder.FileByteStream(car_path)


async def car_bytes(stream, cid: str) -> bytes:
    """
    CAR stream to bytes

    :param stream: CAR stream
    :param str cid: CID of the UnixFS directory to export
    :rtype: bytes
    """

    buff = b''

    async for chunk in car_decoder.stream_bytes(cid, stream):
        buff += chunk

    return buff


class DotJSON(dict):
    """
    Based on edict, described here:

    https://gist.github.com/markhu/fbbab71359af00e527d0
    """

    __delattr__ = dict.__delitem__

    def __init__(self, data):
        if isinstance(data, str):
            data = json.loads(data)
        else:
            data = data

        for name, value in data.items():
            setattr(self, name, self._wrap(value))

    def _traverse(self, obj, attr):
        attrd = attr.replace('__', '-')
        if self._is_indexable(obj):
            try:
                return obj[int(attrd)]
            except:
                return None
        elif isinstance(obj, dict):
            return obj.get(attrd, None)
        else:
            return attrd

    def __getattr__(self, attr):
        if '.' in attr:
            attrVal = attr.split('.').replace('__', '-')
            return reduce(self._traverse, attrVal, self)
        return self.get(attr, None)

    def __setattr__(self, attr, value):
        attrd = attr.replace('__', '-')
        dict.__setitem__(self, attrd, value)

    def _wrap(self, value):
        if self._is_indexable(value):
            # (!) recursive (!)
            return type(value)([self._wrap(v) for v in value])
        elif isinstance(value, dict):
            return DotJSON(value)
        else:
            return value

    @staticmethod
    def _is_indexable(obj):
        return isinstance(obj, (tuple, list, set, frozenset))

    def write(self, fd):
        fd.write(json.dumps(self))
        fd.flush()



import re


class InvalidNodeAddressError(Exception):
    pass


class IPFSConnectionError(Exception):
    pass


class APIError(Exception):
    """
    IPFS API error

    :param int code: IPFS error code
    :param str message: Error message
    :param int http_status: HTTP status code
    """

    def __init__(self, code=-1, message='', http_status=-1):
        self.code = code
        self.message = message
        self.http_status = http_status

    @classmethod
    def match(cls, message: str):
        """
        :param str message: Error message string returned by kubo
        :rtype: bool
        """
        return False


class EndpointNotFoundError(APIError):
    """
    Exception for when a RPC endpoint is not found (HTTP 404)
    """


class NotPinnedError(APIError):
    """
    Content not pinned or pinned indirectly
    """

    @classmethod
    def match(cls, message: str):
        return message.lower() == 'not pinned or pinned indirectly'


class InvalidCIDError(APIError):
    """
    Invalid CID or selected encoding not supported
    """

    @classmethod
    def match(cls, message: str):
        return message.lower().endswith(
            'invalid cid: selected encoding not supported'
        )


class NoSuchLinkError(APIError):
    """
    No link by that name
    """

    @classmethod
    def match(cls, message: str):
        return message == 'no link by that name'


class IpnsKeyError(APIError):
    """
    IPNS key errors
    """

    @classmethod
    def match(cls, message: str):
        ma = re.search(r'^key with name ([\'\\\w]+) already exists',
                       message)
        return ma is not None


class PinRemoteError(APIError):
    """
    Any kind of error ocurring when interacting with a
    remote pinning service.
    """

    @classmethod
    def match(cls, message: str):
        # TODO: match all types of remote pin errors
        return message.startswith(
            'empty response from remote pinning service'
        ) or message.startswith(
            'service endpoint must be a valid HTTP URL'
        ) or 'ERROR_UNRECOGNISED_TOKEN' in message or \
            "ERR_INVALID_TOKEN" in message


class UnknownAPIError(APIError):
    pass


class InvalidPubMessageError(Exception):
    """
    Invalid pubsub message encoding error
    """



import os.path

from yarl import URL
from multiaddr import Multiaddr

from aioipfs.api import SubAPI
from aioipfs.api import ARG_PARAM
from aioipfs.api import boolarg
from aioipfs.exceptions import APIError

from aioipfs.helpers import async_enterable
from aioipfs.helpers import maddr_tcp_explode
from aioipfs.helpers import quote_args
from aioipfs.helpers import quote_dict
from aioipfs.helpers import p2p_addr_explode


def protocol_format(proto: str, ns='x') -> str:
    if proto.startswith('/x/'):
        return proto

    return f'/{ns}/{proto}'


class TunnelDialerContext:
    """
    Service dial context manager
    """

    def __init__(self,
                 client,
                 peerId: str,
                 proto: str,
                 maddr: Multiaddr):
        self.client = client
        self.peerId = peerId
        self.protocol = proto
        self.maddr = maddr

    @property
    def maddr_port(self):
        """
        The connection TCP port
        """
        if self.maddr:
            return maddr_tcp_explode(self.maddr)[1]

    @property
    def maddr_host(self):
        """
        The connection hostname/ip
        """
        if self.maddr:
            return maddr_tcp_explode(self.maddr)[0]

    @property
    def failed(self):
        """
        Returns True if the service dial has failed
        """
        return self.maddr is None

    def httpUrl(self,
                path: str,
                query: str = '',
                fragment: str = '',
                scheme: str = 'http') -> URL:
        """
        Returns an HTTP URL targeting the remote P2P service
        associated with this context.

        :param str path: HTTP path
        :param str query: HTTP query
        :param str fragment: HTTP fragment
        :param str scheme: URL scheme (http or https)
        :rtype: URL
        """
        return URL.build(
            host=self.maddr_host,
            port=self.maddr_port,
            scheme=scheme,
            path=path,
            query=query,
            fragment=fragment
        )

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        try:
            assert await self.client.p2p.listeners_for_listenaddr(
                self.maddr
            )
        except Exception:
            pass
        else:
            await self.client.p2p.listener_close(
                self.protocol,
                listen_address=str(self.maddr)
            )


class P2PDialerMixin:
    """
    APIs to help with dialing remote p2p services
    """

    @async_enterable
    async def dial_service(self,
                           peer: str,
                           protocol: str,
                           address=None,
                           auto=True,
                           allow_loopback=False):
        """
        Dial a remote IPFS p2p service.

        Returns an async context manager:

        async with client.p2p.dial_service(...) as ctx:
            print(ctx.maddr_host, ctx.maddr_port)

        :param str peer: Peer ID
        :param str protocol: Service protocol (e.g /x/helloworld)
        :param str address: Multiaddr for the connection
        :param bool auto: Automatically assign a multiaddr (only works
            if the kubo daemon is local)
        :param bool allow_loopback: Allow dialing a service on this node
        """
        if auto is True and not address and self.driver.host_local:
            address = self.driver.allocate_tcp_maddr()

        if await self.driver.agent_version_post0418():
            proto = protocol_format(protocol)
        else:
            proto = protocol

        return await self._dial_start(peer, proto, address,
                                      allow_loopback=allow_loopback)

    @async_enterable
    async def dial_endpoint(self,
                            p2pEndpointAddr: str,
                            address=None,
                            auto=True,
                            allow_loopback=False):
        """
        Dial a remote p2p service by passing an endpoint address.

        /p2p/12D3KooWAci7T5tA1ZaxBBREVEdLX5FJkH2GBcgbJGW6LqJtBgOp/x/hello

        Returns an async context manager:

        async with client.p2p.dial_endpoint('/p2p/...') as ctx:
            ...

        :param str p2pEndpointAddr: Endpoint address
        :param str address: Multiaddr for the connection
        :param bool auto: Automatically assign a multiaddr (only works
            if the kubo daemon is local)
        :param bool allow_loopback: Allow dialing a service on this node
        """

        if auto is True and not address and self.driver.host_local:
            address = self.driver.allocate_tcp_maddr()

        exploded = p2p_addr_explode(p2pEndpointAddr)

        if exploded:
            peerId, protoFull, pVersion = exploded

            return await self._dial_start(
                peerId, protoFull, str(address),
                allow_loopback=allow_loopback
            )
        else:
            raise ValueError(f'Invalid P2P service address: {p2pEndpointAddr}')

    async def _dial_start(self,
                          peer: str,
                          proto: str,
                          address: str,
                          allow_loopback=False):
        """
        Start the P2P dial and return a TunnelDialerContext
        """

        if not address:
            raise ValueError('No multiaddr provided')

        node_id = (await self.driver.core.id())['ID']

        if allow_loopback is True and node_id == peer:
            listeners = await self.listeners_for_proto(proto)

            for listener in listeners:
                # Return the first matching ListenAddress

                if listener['ListenAddress'] == f'/p2p/{node_id}':
                    return TunnelDialerContext(
                        self.driver, peer, proto,
                        Multiaddr(listener['TargetAddress'])
                    )

        try:
            peerAddr = os.path.join(
                '/ipfs', peer) if not peer.startswith('/ipfs/') else peer
            resp = await self.driver.p2p.dial(proto, address, peerAddr)

            if resp:
                maddr = Multiaddr(resp['Address'])
            elif isinstance(address, Multiaddr):
                maddr = address
            elif isinstance(address, str):
                maddr = Multiaddr(address)
        except Exception:
            return TunnelDialerContext(self.driver, peer, proto, None)

        return TunnelDialerContext(self.driver, peer, proto, maddr)

    async def streams_for_proto(self, protocol: str) -> list:
        return [stream for stream in await self.streams() if
                stream['Protocol'] == protocol]

    async def streams_for_targetaddr(self, addr: Multiaddr) -> list:
        return [stream for stream in await self.streams() if
                stream['TargetAddress'] == str(addr)]

    async def listeners_for_targetaddr(self, addr: Multiaddr) -> list:
        return [stream for stream in await self.listeners() if
                stream['TargetAddress'] == str(addr)]

    async def listeners_for_listenaddr(self, addr: Multiaddr) -> list:
        return [stream for stream in await self.listeners() if
                stream['ListenAddress'] == str(addr)]

    async def listeners_for_proto(self, protocol: str) -> list:
        return [stream for stream in await self.listeners() if
                stream['Protocol'] == protocol]

    async def listeners(self) -> list:
        try:
            resp = await self.ls(headers=True)
            assert isinstance(resp, dict)
            return resp['Listeners'] if resp['Listeners'] else []
        except (AssertionError, APIError):
            return []

    async def streams(self) -> list:
        try:
            resp = await self.stream_ls(headers=True)
            assert isinstance(resp, dict)
            return resp['Streams'] if resp['Streams'] else []
        except (AssertionError, APIError):
            return []


class P2PAPI(P2PDialerMixin, SubAPI):
    """
    P2P API.

    Note: go-ipfs v0.4.18 introduced some changes in the P2P subsystem
    and some endpoints were renamed. Agent version detection is done
    in the affected API calls to maintain compatibility.
    """
    async def listener_open(self, protocol, address,
                            allow_custom_protocol=False,
                            report_peerid=False):
        """
        Open a P2P listener

        :param str protocol: protocol name associated with the listener
        :param str address: address for the listener, in multiaddr format
        """

        api_post0418 = await self.driver.agent_version_post0418()

        if api_post0418 is True:
            url = self.url('p2p/listen')
            params = quote_dict({
                ARG_PARAM: [protocol, address],
                'allow-custom-protocol': boolarg(allow_custom_protocol),
                'report-peer-id': boolarg(report_peerid)
            })
        else:
            url = self.url('p2p/listener/open')
            params = quote_args(protocol, address)

        return await self.fetch_json(url, params=params)

    async def listener_close(self, protocol, listen_address=None,
                             target_address=None, all=False):
        """
        Close a previously opened P2P listener

        :param str protocol: protocol name associated with the listener
        :param bool all: if True, closes all listeners on the node
        """

        api_post0418 = await self.driver.agent_version_post0418()

        if api_post0418 is True:
            url = self.url('p2p/close')
            params = {
                'protocol': protocol,
                'all': boolarg(all)
            }

            if listen_address:
                params['listen-address'] = listen_address
            if target_address:
                params['target-address'] = target_address
        else:
            url = self.url('p2p/listener/close')
            params = {
                ARG_PARAM: protocol,
                'all': boolarg(all)
            }

        return await self.fetch_json(url, params=params)

    async def listener_ls(self, headers=False):
        """
        List P2P listeners

        :param bool headers: print all headers (HandlerID, Protocol, ...)
        """

        api_post0418 = await self.driver.agent_version_post0418()

        if api_post0418 is True:
            url = self.url('p2p/ls')
        else:
            url = self.url('p2p/listener/ls')

        return await self.fetch_json(url,
                                     params={'headers': boolarg(headers)})

    async def stream_dial(self, protocol, laddress, target,
                          allow_custom_protocol=False):
        """
        Dial to a P2P listener.

        :param str peer: Remote Peer ID
        :param str protocol: protocol identifier
        :param str address: multiaddr to listen for connection/s
            (default: /ip4/127.0.0.1/tcp/0)
        """

        api_post0418 = await self.driver.agent_version_post0418()

        if api_post0418 is True:
            url = self.url('p2p/forward')
            params = quote_dict({
                ARG_PARAM: [protocol, laddress, target],
                'allow-custom-protocol': boolarg(allow_custom_protocol)
            })
        else:
            url = self.url('p2p/stream/dial')
            args = [
                target,
                protocol,
                laddress] if laddress else [
                target,
                protocol]
            params = quote_args(*args)

        return await self.fetch_json(url, params=params)

    async def stream_close(self, stream_id=None, all=False):
        """
        Close active P2P stream.
        """

        if stream_id and all is False:
            params = {
                ARG_PARAM: stream_id,
            }
        elif all is True:
            params = {
                'all': boolarg(all)
            }
        else:
            raise ValueError('Invalid stream parameters')

        return await self.fetch_text(self.url('p2p/stream/close'),
                                     params=params)

    async def stream_ls(self, headers=False):
        """
        List active P2P streams.
        """
        return await self.fetch_json(self.url('p2p/stream/ls'),
                                     params={'headers': boolarg(headers)})

    listen = listener_open
    dial = stream_dial
    forward = stream_dial
    ls = listener_ls



from aioipfs.api import SubAPI
from aioipfs.api import boolarg
from aioipfs import multi


class MultibaseAPI(SubAPI):
    async def list(self, prefix=False, numeric=False):
        """
        List available multibase encodings
        """

        return await self.fetch_json(
            self.url('multibase/list'),
            params={
                'prefix': boolarg(prefix),
                'numeric': boolarg(numeric)
            }
        )

    async def decode(self, filepath: str):
        """
        Decode multibase string

        :param str filepath: Path to the file containing the data to encode
        """

        with multi.FormDataWriter() as mpwriter:
            mpwriter.append_payload(multi.bytes_payload_from_file(filepath))

            return await self.post(self.url('multibase/decode'),
                                   mpwriter, outformat='text')

    async def __encode_transcode(self, method: str,
                                 filepath: str, base='base64url'):
        """
        Method used for encoding or transcoding
        """

        params = {
            'b': base
        }

        with multi.FormDataWriter() as mpwriter:
            mpwriter.append_payload(multi.bytes_payload_from_file(filepath))

            return await self.post(self.url(f'multibase/{method}'),
                                   mpwriter, params=params, outformat='text')

    async def encode(self, filepath: str, base='base64url'):
        """
        Encode data into multibase string

        :param str filepath: Path to the file containing the data to encode
        :param str base: multibase encoding. Default: base64url
        """
        return await self.__encode_transcode('encode', filepath, base=base)

    async def transcode(self, filepath: str, base='base64url'):
        """
        Transcode multibase string between bases

        :param str filepath: Path to the file containing the data to transcode
        :param str base: multibase encoding. Default: base64url
        """
        return await self.__encode_transcode('transcode', filepath, base=base)



from aioipfs.apis import SubAPI
from aioipfs.helpers import *  # noqa


class PinRemoteServiceAPI(SubAPI):
    async def add(self, service, endpoint, key):
        """
        Add remote pinning service.

        :param str service: Service name
        :param str endpoint: Service endpoint
        :param str key: Service key
        """
        return await self.fetch_json(
            self.url('pin/remote/service/add'),
            params=quote_args(service, endpoint, key))

    async def ls(self, stat=False):
        """
        List remote pinning services.

        :param bool stat: Try to fetch and display current pin count
            on remote service (queued/pinning/pinned/failed)
        """
        params = {
            'stat': boolarg(stat)
        }

        return await self.fetch_json(self.url('pin/remote/service/ls'),
                                     params=params)

    async def rm(self, service):
        """
        Remove remote pinning service.

        :param str service: Service name
        """
        return await self.fetch_json(
            self.url('pin/remote/service/rm'),
            params=quote_args(service))


class PinRemoteAPI(SubAPI):
    def __init__(self, driver):
        super().__init__(driver)

        self.service = PinRemoteServiceAPI(driver)

    async def add(self,
                  service: str, objPath: str,
                  name=None,
                  background=False):
        """
        Pin object to remote pinning service.

        :param str service: Service name
        :param str objPath: Path to object(s) to be pinned
        :param str name: An optional name for the pin
        :param bool background: Add to the queue on the remote service
            and return immediately (does not wait for pinned status)
            Default: false
        """

        params = {
            'service': service,
            ARG_PARAM: objPath,
            'background': boolarg(background)
        }

        if isinstance(name, str):
            params['name'] = name

        return await self.fetch_json(
            self.url('pin/remote/add'),
            params=params
        )

    async def ls(self, service: str,
                 name=None,
                 cid: list = [],
                 status: list = ['pinned']):
        """
        List objects pinned to remote pinning service.

        :param str service: Name of the remote pinning service to use
        :param str name: Return pins with names that contain the value
            provided (case-sensitive, exact match)
        :param list cid: Return pins for the specified CIDs
        :param list status: Return pins with the specified statuses
            (queued,pinning,pinned,failed). Default: [pinned]
        """

        params = {
            'service': service,
            'status': status
        }

        if len(cid) > 0:
            params['cid'] = cid

        if isinstance(name, str):
            params['name'] = name

        async for entry in self.mjson_decode(
                self.url('pin/remote/ls'),
                params=quote_dict(params)):
            yield entry

    async def rm(self, service: str,
                 name=None,
                 cid: list = [],
                 status: list = ['pinned'],
                 force=False):
        """
        Remove pins from remote pinning service.

        :param str service: Name of the remote pinning service to use
        :param str name: Return pins with names that contain the value
            provided (case-sensitive, exact match)
        :param list cid: Return pins for the specified CIDs
        :param list status: Return pins with the specified statuses
            (queued,pinning,pinned,failed). Default: [pinned]
        :param bool force: Allow removal of multiple pins matching the
            query without additional confirmation. Default: false
        """
        params = {
            'service': service,
            'status': status,
            'force': boolarg(force)
        }

        if len(cid) > 0:
            params['cid'] = cid

        if isinstance(name, str):
            params['name'] = name

        return await self.fetch_json(self.url('pin/remote/rm'),
                                     params=quote_dict(params))


class PinAPI(SubAPI):
    def __init__(self, driver):
        super().__init__(driver)

        self.remote = PinRemoteAPI(driver)

    async def add(self, path, recursive=True, progress=True):
        """
        Pin objects to local storage.

        :param str path: Path to object(s) to be pinned
        :param bool recursive: Recursively pin the object linked to
            by the specified object(s)
        :param bool progress: Show progress
        """

        # We request progress status by default
        params = {
            ARG_PARAM: path,
            'recursive': boolarg(recursive),
            'progress': boolarg(progress)
        }

        async for added in self.mjson_decode(
                self.url('pin/add'), params=params):
            yield added

    async def ls(self, path=None, pintype='all', quiet=False,
                 stream: bool = False):
        """
        List objects pinned to local storage.

        :param str path: Path of object(s) to be listed
        :param str pintype: The type of pinned keys to list. Can be
            "direct", "indirect", "recursive", or "all"
        :param bool quiet: Write just hashes of objects
        :param bool stream: Enable streaming of pins as they are discovered
        """

        params = {
            'type': pintype,
            'quiet': boolarg(quiet),
            'stream': boolarg(stream)
        }
        if path:
            params[ARG_PARAM] = path

        return await self.fetch_json(self.url('pin/ls'),
                                     params=params)

    async def rm(self, path, recursive=True):
        """
        Remove pinned objects from local storage.

        :param str path: Path of object(s) to be removed
        :param bool recursive: Recursively unpin the object linked to
            by the specified object(s)
        """

        params = {
            ARG_PARAM: path,
            'recursive': boolarg(recursive)
        }
        return await self.fetch_json(self.url('pin/rm'),
                                     params=params)

    async def verify(self, verbose=False, quiet=True):
        """
        Verify that recursive pins are complete.

        :param bool verbose: Also write the hashes of non-broken pins
        :param bool quiet: Write just hashes of broken pins
        """

        params = {
            'verbose': boolarg(verbose),
            'quiet': boolarg(quiet)
        }
        return await self.fetch_json(self.url('pin/verify'),
                                     params=params)

    async def update(self, old, new, unpin=True):
        """
        Update a recursive pin

        :param str old: Path to old object
        :param str new: Path to new object to be pinned
        :param bool unpin: Remove the old pin
        """

        params = quote_dict({
            ARG_PARAM: [old, new],
            'unpin': boolarg(unpin)
        })
        return await self.fetch_json(self.url('pin/update'), params=params)



import functools
import json
import re
import sys
from multiaddr import Multiaddr
from distutils.version import StrictVersion
from contextlib import closing
import socket
import os.path

from urllib.parse import quote

from multiformats_cid import make_cid
from multiformats_cid import CIDv1
from multiformats_cid import CIDv0


try:
    import orjson
except ImportError:
    have_orjson = False
else:
    have_orjson = True


ARG_PARAM = 'arg'


def boolarg(arg):
    return str(arg).lower()


def quote_args(*args):
    # Used in the few cases where there are multiple 'arg=' URL params
    # that yarl can't handle at the moment
    quoted = ''

    if len(args) > 0:
        for arg in args:
            quoted += '&{0}={1}'.format(ARG_PARAM, quote(str(arg)))
        return quoted[1:]


def quote_dict(data):
    quoted = ''

    if not isinstance(data, dict):
        raise ValueError('quote_dict: need dictionary')

    for arg, value in data.items():
        if isinstance(value, list):
            for lvalue in value:
                quoted += '&{0}={1}'.format(arg, quote(str(lvalue)))
        elif isinstance(value, str):
            quoted += '&{0}={1}'.format(arg, quote(value))
        elif isinstance(value, int):
            quoted += '&{0}={1}'.format(arg, quote(str(value)))
        elif isinstance(value, bool):
            quoted += '&{0}={1}'.format(arg, quote(boolarg(value)))

    if len(quoted) > 0:
        return quoted[1:]


def decode_json(data: bytes):
    try:
        if not isinstance(data, bytes):
            raise TypeError('decode_json: invalid value type')

        if have_orjson:
            return orjson.loads(data.decode())
        else:
            return json.loads(data.decode())
    except Exception as exc:
        print('Could not read JSON object:', str(exc), file=sys.stderr)
        return None


def async_enterable(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        class AsyncEnterableInstance:
            async def __aenter__(self):
                self.context = await f(*args, **kwargs)
                return await self.context.__aenter__()

            async def __aexit__(self, *args, **kwargs):
                await self.context.__aexit__(*args, **kwargs)

            def __await__(self):
                return f(*args, **kwargs).__await__()

        return AsyncEnterableInstance()

    return wrapper


def maddr_tcp_explode(multi: Multiaddr) -> tuple:
    host, port = None, 0

    for proto in multi.protocols():
        if proto.name in ['ip4', 'ip6', 'dns4', 'dns6']:
            host = multi.value_for_protocol(proto.code)
        if proto.name == 'tcp':
            port = int(multi.value_for_protocol(proto.code))

    return host, port


def unusedTcpPort():
    try:
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            s.bind(('', 0))
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            return s.getsockname()[1]
    except Exception:
        return None


def p2p_addr_explode(addr: str) -> tuple:
    """
    Explode a P2P service endpoint address such as :

    /p2p/12D3KooWD3bfmNbuuuVCYwkjnFt3ukm3qaB3hDED3peHHXawvRAi/x/videocall/room1/1.0.0
    /p2p/12D3KooWD3bfmNbuuuVCYwkjnFt3ukm3qaB3hDED3peHHXawvRAi/x/test

    into its components, returning a tuple in the form

    (peerId, protoFull, protoVersion)

    protoFull can be passed to 'ipfs p2p dial'
    """

    peerIdRe = re.compile(r'([\w]){46,59}$')

    parts = addr.lstrip(os.path.sep).split(os.path.sep)
    try:
        assert parts.pop(0) == 'p2p'
        peerId = parts.pop(0)
        prefix = parts.pop(0)
        assert peerIdRe.match(peerId)
        assert prefix in ['x', 'y', 'z']

        pVersion = None
        protoA = [prefix]
        protoPart = parts.pop(0)
        protoA.append(protoPart)

        while protoPart:
            try:
                protoPart = parts.pop(0)
            except IndexError:
                break

            protoA.append(protoPart)

            try:
                v = StrictVersion(protoPart)
            except Exception:
                # No version
                pass
            else:
                # Found a version, should be last element
                pVersion = v
                assert len(parts) == 0
                break

        return peerId, os.path.sep + os.path.join(*protoA), pVersion
    except Exception:
        return None


def peerid_reencode(peerId: str,
                    base: str = 'base36',
                    multicodec: str = 'libp2p-key') -> str:
    """
    Encode a PeerId to a specific base
    """

    cid = make_cid(peerId)
    if not cid:
        return None

    if base in ['base32', 'base36']:
        return CIDv1(multicodec, cid.multihash).encode(base).decode()
    elif base in ['base58']:
        return str(CIDv0(cid.multihash))

    return None


def peerid_base32(peerId: str) -> str:
    """
    Convert any PeerId to a CIDv1 (base32)
    """
    return peerid_reencode(peerId, base='base32')


def peerid_base36(peerId: str) -> str:
    """
    Convert any PeerId to a CIDv1 (base36)
    """
    return peerid_reencode(peerId, base='base36')


def peerid_base58(peerId: str) -> str:
    """
    Convert any PeerId to a CIDv0
    """
    return peerid_reencode(peerId, base='base58')



import os
import os.path
import posixpath
import io
import re

from gitignore_parser import parse_gitignore
import aiohttp
from aiohttp import payload


class FormDataWriter(aiohttp.MultipartWriter):
    def __init__(self):
        super().__init__(subtype='form-data')


def multiform_bytes(bytes, name=''):
    with FormDataWriter() as mpwriter:
        part = payload.BytesPayload(
            bytes, content_type='application/octet-stream'
        )
        part.set_content_disposition('form-data',
                                     name='file', filename=name)
        mpwriter.append_payload(part)
        return mpwriter


def multiform_json(json, name=''):
    with FormDataWriter() as mpwriter:
        part = payload.JsonPayload(json)
        part.set_content_disposition('form-data',
                                     name='file', filename=name)
        mpwriter.append_payload(part)
        return mpwriter


def bytes_payload_from_file(filepath):
    basename = os.path.basename(filepath)
    file_payload = payload.BytesIOPayload(
        open(filepath, 'rb'),
        content_type='application/octet-stream'
    )

    file_payload.set_content_disposition('form-data',
                                         name='file',
                                         filename=basename)
    return file_payload


def glob_compile(pat):
    """ From ipfsapi.multipart

       Translate a shell glob PATTERN to a regular expression.

    This is almost entirely based on `fnmatch.translate` source-code from the
    python 3.5 standard-library.
    """

    i, n = 0, len(pat)
    res = ''
    while i < n:
        c = pat[i]
        i = i + 1
        if c == '/' and len(pat) > (i + 2) and pat[i:(i + 3)] == '**/':
            # Special-case for "any number of sub-directories" operator since
            # may also expand to no entries:
            #  Otherwise `a/**/b` would expand to `a[/].*[/]b` which wouldn't
            #  match the immediate sub-directories of `a`, like `a/b`.
            i = i + 3
            res = res + '[/]([^/]*[/])*'
        elif c == '*':
            if len(pat) > i and pat[i] == '*':
                i = i + 1
                res = res + '.*'
            else:
                res = res + '[^/]*'
        elif c == '?':
            res = res + '[^/]'
        elif c == '[':
            j = i
            if j < n and pat[j] == '!':
                j = j + 1
            if j < n and pat[j] == ']':
                j = j + 1
            while j < n and pat[j] != ']':
                j = j + 1
            if j >= n:
                res = res + '\\['
            else:
                stuff = pat[i:j].replace('\\', '\\\\')
                i = j + 1
                if stuff[0] == '!':
                    stuff = '^' + stuff[1:]
                elif stuff[0] == '^':
                    stuff = '\\' + stuff
                res = '%s[%s]' % (res, stuff)
        else:
            res = res + re.escape(c)
    return re.compile(r'(?ms)^' + res + r'\Z' + '$')


class DirectoryListing:
    """ This class is from ipfsapi.multipart.DirectoryStream, reused it
        just to generate the file paths. It's not a generator """

    def __init__(self,
                 directory,
                 recursive=False,
                 patterns='**',
                 hidden=False,
                 ignrulespath=None,
                 chunk_size=4096):

        self.hidden = hidden
        self.ignrulespath = ignrulespath
        self.ign_path = None
        self.ign_short_name = None
        self.ignrules = None
        self.patterns = []
        patterns = [patterns] if isinstance(patterns, str) else patterns
        for pattern in patterns:
            if isinstance(pattern, str):
                self.patterns.append(glob_compile(pattern))
            else:
                self.patterns.append(pattern)

        self.directory = os.path.normpath(directory)
        self.recursive = recursive

        if self.ignrulespath:
            self.ign_short_name = os.path.join(
                os.path.basename(self.directory), self.ignrulespath)
            self.ign_path = os.path.join(self.directory, self.ignrulespath)

    def genNames(self):
        """ Returns the file paths inside self.directory
            with associated opened file descriptors """
        names = []

        added_directories = set()

        def strip_first(path):
            try:
                return os.path.sep.join(path.strip(os.path.sep).split(
                    os.path.sep)[1:])
            except Exception:
                return None

        def hidden_ignore(fpath):
            if not self.hidden:
                return any(
                    p.startswith('.') for p in fpath.split(os.path.sep)
                )

            return False

        def add_directory(short_path):
            # Do not continue if this directory has already been added
            if hidden_ignore(short_path):
                return

            dir_unr = strip_first(short_path)
            dir_fpath = os.path.join(self.directory, dir_unr)

            if dir_unr and self.ignrules and self.ignrules(dir_fpath):
                # Matches ignore rules
                return

            if short_path in added_directories:
                return

            # Scan for first super-directory that has already been added
            dir_base = short_path
            dir_parts = []
            while dir_base:
                dir_base, dir_name = os.path.split(dir_base)

                dir_parts.append(dir_name)
                if dir_base in added_directories:
                    break

            # Add missing intermediate directory nodes in the right order
            while dir_parts:
                dir_base = os.path.join(dir_base, dir_parts.pop())

                # Create an empty, fake file to represent the directory
                mock_file = io.StringIO()
                mock_file.write(u'')

                # posix-ify
                dir_base = dir_base.replace(os.sep, posixpath.sep)

                # Add this directory to those that will be sent
                names.append(('files',
                              (dir_base,
                               mock_file,
                               'application/x-directory')))
                # Remember that this directory has already been sent
                added_directories.add(dir_base)

        def add_file(short_path, full_path, force=False):
            try:
                if not full_path == self.ign_path and \
                        hidden_ignore(short_path):
                    # Ignore hidden files
                    return

                short_unr = strip_first(short_path)
                if short_unr and self.ignrules and self.ignrules(full_path):
                    # Matches ignore rules
                    return

                # posix-ify
                short_path = short_path.replace(os.sep, posixpath.sep)

                # Always add files in wildcard directories
                names.append(('files', (short_path,
                                        open(full_path, 'rb'),
                                        'application/octet-stream')))
            except OSError:
                # File might have disappeared between `os.walk()` and `open()`
                pass

        def match_short_path(short_path):
            # Remove initial path component so that all files are based in
            # the target directory itself (not one level above)
            if os.sep in short_path:
                path = short_path.split(os.sep, 1)[1]
            else:
                return False

            # Convert all path seperators to POSIX style
            path = path.replace(os.sep, '/')

            # Do the matching and the simplified path
            for pattern in self.patterns:
                if pattern.match(path):
                    return True
            return False

        # Identify the unecessary portion of the relative path
        truncate = os.path.dirname(self.directory)
        # Traverse the filesystem downward from the target directory's uri
        # Errors: `os.walk()` will simply return an empty generator if the
        #         target directory does not exist.
        wildcard_directories = set()

        # Support ignore rules at folder's root
        if self.ign_path:
            if os.path.exists(self.ign_path) and not self.ignrules:
                try:
                    self.ignrules = parse_gitignore(
                        self.ign_path,
                        self.directory
                    )

                except Exception:
                    self.ignrules = None

        for curr_dir, _, files in os.walk(self.directory):
            # find the path relative to the directory being added
            if len(truncate) > 0:
                _, _, short_path = curr_dir.partition(truncate)
            else:
                short_path = curr_dir
            # remove leading / or \ if it is present
            if short_path.startswith(os.sep):
                short_path = short_path[1:]

            wildcard_directory = False
            if os.path.split(short_path)[0] in wildcard_directories:
                # Parent directory has matched a pattern, all sub-nodes should
                # be added too
                wildcard_directories.add(short_path)
                wildcard_directory = True
            else:
                # Check if directory path matches one of the patterns
                if match_short_path(short_path):
                    # Directory matched pattern and it should therefor
                    # be added along with all of its contents
                    wildcard_directories.add(short_path)
                    wildcard_directory = True

            # Always add directories within wildcard directories - even if they
            # are empty
            if wildcard_directory:
                add_directory(short_path)

            # Iterate across the files in the current directory
            for filename in files:
                # Find the filename relative to the directory being added

                short_name = os.path.join(short_path, filename)
                filepath = os.path.join(curr_dir, filename)

                if wildcard_directory:
                    # Always add files in wildcard directories
                    add_file(short_name, filepath)
                else:
                    # Add file (and all missing intermediary directories)
                    # if it matches one of the patterns
                    if match_short_path(short_name):
                        add_directory(short_path)
                        add_file(short_name, filepath)

        return names



import io
import asyncio
from typing import Union

import base58
import base64
import multibase

from aiohttp import payload
from aioipfs import multi
from aioipfs.helpers import quote_args
from aioipfs.api import SubAPI
from aioipfs.api import ARG_PARAM
from aioipfs.api import boolarg
from aioipfs.exceptions import InvalidPubMessageError


class PubSubAPI(SubAPI):
    async def ls(self):
        """
        List the names of the subscribed pubsub topics.

        :rtype dict:
        """

        api_post011 = await self.driver.agent_version_post011()

        reply = await self.fetch_json(self.url('pubsub/ls'))

        if not isinstance(reply, dict) or 'Strings' not in reply:
            return None

        if api_post011:
            # Do the multibase decoding of the topics list
            return {
                'Strings': [multibase.decode(topic).decode() for topic in
                            reply['Strings']]
            }
        else:
            return reply

    async def peers(self, topic=None):
        """
        List peers communicating over pubsub with this node.
        """

        params = {}
        if isinstance(topic, str):
            params[ARG_PARAM] = topic

        return await self.fetch_json(
            self.url('pubsub/peers'), params=params)

    async def pub(self, topic, data: Union[str, bytes]):
        """
        Publish a message to a given pubsub topic.

        :param str topic: topic to publish the message to
        :param data: message data (bytes or str)
        """

        api_post011 = await self.driver.agent_version_post011()

        if type(data) not in [str, bytes]:
            raise ValueError(
                'Invalid data type (should be bytes or str)'
            )

        if api_post011:
            return await self.__pub_multibase(topic, data)
        else:
            return await self.__pub_old(topic, data)

    async def __pub_old(self, topic, data: Union[str, bytes]):
        """
        Old publish wire format (pre 0.11.0)
        """

        msgd = data.decode() if isinstance(data, bytes) else data

        return await self.post(self.url('pubsub/pub'),
                               None, params=quote_args(topic, msgd))

    async def __pub_multibase(self, topic, data: Union[str, bytes]):
        """
        New publish wire format (post 0.11.0)
        """
        params = {
            ARG_PARAM: multibase.encode('base64url', topic).decode()
        }

        with multi.FormDataWriter() as mpwriter:
            if isinstance(data, str):
                part = payload.StringIOPayload(io.StringIO(data))
            elif isinstance(data, bytes):
                part = payload.BytesIOPayload(io.BytesIO(data))

            part.set_content_disposition('form-data', name='file')
            mpwriter.append_payload(part)

            return await self.post(self.url('pubsub/pub'), mpwriter,
                                   params=params, outformat='text')

    async def sub(self,
                  topic: str,
                  discover: bool = True,
                  timeout: int = None,
                  read_timeout: int = None):
        """
        Subscribe to messages on a given topic.

        This is an async generator yielding messages as they are read on the
        pubsub topic.

        :param str topic: topic to subscribe to
        :param bool discover: try to discover other peers subscribed to
            the same topic
        :param int timeout: global coroutine timeout
        :param int read_timeout: read timeout (between each message read)
        """

        api_post011 = await self.driver.agent_version_post011()

        params = {
            'discover': boolarg(discover),
            'stream-channels': boolarg(True)
        }

        if api_post011:
            # post v0.11.0: topic is encoded with multibase
            params[ARG_PARAM] = multibase.encode(
                'base64url', topic).decode()
        else:
            # pre v0.11.0: topic is passed as a raw string
            params[ARG_PARAM] = topic

        async for message in self.mjson_decode(
                self.url('pubsub/sub'),
                method='post',
                headers={'Connection': 'Close'},
                new_session=True,
                timeout=timeout if timeout else 60.0 * 60 * 24 * 8,
                read_timeout=read_timeout if read_timeout else 0,
                params=params):
            try:
                if api_post011:
                    converted = self._decode_message(message)
                else:
                    converted = self._decode_message_base58(message)
            except Exception as e:
                raise InvalidPubMessageError(
                    f'Could not decode pubsub message '
                    f'on topic {topic}: {e}'
                )
            else:
                yield converted

            await asyncio.sleep(0)

    def _decode_message(self, psmsg):
        """
        Starting with go-ipfs v0.11, JSON fields in the pubsub messages
        are simply multibase-encoded (except the 'from' field which
        represents the PeerID)

        This is the wire format used by go-ipfs/kubo >= 0.11.0
        """

        conv_msg = {}
        conv_msg['from'] = psmsg['from']
        conv_msg['data'] = multibase.decode(psmsg['data'])
        conv_msg['seqno'] = multibase.decode(psmsg['seqno'])
        conv_msg['topicIDs'] = [multibase.decode(topic).decode() for topic in
                                psmsg['topicIDs']]
        return conv_msg

    def _decode_message_base58(self, psmsg):
        """
        Convert a raw pubsub message (with base64-encoded fields) to a
        readable form

        This is the wire format used by go-ipfs/kubo < 0.11.0
        """

        conv_msg = {}
        conv_msg['from'] = base58.b58encode(
            base64.b64decode(psmsg['from']))
        conv_msg['data'] = base64.b64decode(psmsg['data'])
        conv_msg['seqno'] = base64.b64decode(psmsg['seqno'])
        conv_msg['topicIDs'] = psmsg['topicIDs']
        return conv_msg



import os
import os.path
import re
import sys
import codecs

from setuptools import setup
from setuptools import find_packages

if sys.version_info < (3, 6):
    raise RuntimeError("You need python3.6 or newer")

with codecs.open(os.path.join(os.path.abspath(os.path.dirname(
        __file__)), 'aioipfs', '__init__.py'), 'r', 'latin1') as fp:
    try:
        version = re.findall(r"^__version__ = '([^']+)'\r?$",
                             fp.read(), re.M)[0]
    except IndexError:
        raise RuntimeError('Unable to determine version.')

with open("README.rst", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    install_reqs = f.read().splitlines()

with open('requirements-dev.txt') as f:
    install_test_reqs = f.read().splitlines()

setup(
    name='aioipfs',
    version=version,
    license='LGPLv3',
    author='cipres',
    author_email='alkaline@gmx.co.uk',
    url='https://gitlab.com/cipres/aioipfs',
    description='Asynchronous IPFS client library',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    packages=find_packages(exclude=['tests']),
    include_package_data=False,
    install_requires=install_reqs,
    extras_require={
        'orjson': ['orjson>=3.0'],
        'car': ['ipfs-car-decoder==0.1.1'],
        'test': install_test_reqs
    },
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Intended Audience :: Developers',
        'Development Status :: 5 - Production/Stable',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',  # noqa
        'Topic :: System :: Filesystems',
        'Topic :: Communications :: File Sharing'
    ],
    keywords=[
        'asyncio',
        'aiohttp',
        'ipfs'
    ]
)
